/*
	Read the template file as chunks
*/

/*
	From the ruby:
		$x.x.x - looks up in the hash, can recurse
		loop x - foreach in array, can recurse
		$x     - just the string
*/

static import rx = std.regexp;
import std.string;



/*
	{#if var}
	{#nif var}
	{#endif}
	{#else}
	{#loop var as v}
	{#endloop}
	{#include file}
	{$var}




	Some special vars:
	std.* - these are always there
	inc.* - given in the include tag
*/


class Template {
	// Variables can be an associative array, a regular array, or a string.
	// Or an integer or a bool. Or a HTMLString.

struct TemplateVariable {
	enum Type { String, Hash, Array }
	Type t;

	//union {
		string str;
		TemplateVariable[] arr;
		TemplateVariable[string] hash;
	//}
/*
	void opAssign(string s){
		t = Type.String;
		str = s;
	}

	void opAssign(TemplateVariable tv) {
		t = tv.t;
		switch(t){
		  case Type.String:
			str = tv.str;
		  break;
		  case Type.Hash:
		  	hash = tv.hash;
		  break;
		  case Type.Array:
		  	arr = tv.arr;
		  break;
		}
	}
*/

	string toString(){
		if(t != Type.String)
			throw new Exception("Template error: type mismatch");
		return str;
	}
}
struct Block {
	enum Type { If, Loop, Normal, Include, Variable }
	struct IfBlock {
		string variable;
		bool negated; // true if nif
		Block* trueBlock;
		Block* falseBlock; // the else
	}
	struct LoopBlock {
		string variable;
		string as;
		Block* contents;
	}
	struct NormalBlock {
		string contents;
	}
	struct IncludeBlock {
		string[string] variables;
		Block* contents;
	}
	struct VariableBlock {
		string variable;
	}

	Type t;
	union {
		IfBlock ifb;
		LoopBlock lb;
		NormalBlock nb;
		IncludeBlock ib;
		VariableBlock vb;
	}

	Block* next;
}


	// terminating 0 is end of file, 1 is else or endif, 2 is endif, 3 is endloop.
Block* readTemplate(string t, int* ending = null, int terminating = 0){
	Block* start = null;
	Block* current = null;
	int numElements = 0;

	void appendBlock(Block* b){
		if(start is null) {
			start = b;
		} else {
			current.next = b;
		}

		current = b;
	}

	auto s = t;
	int abspos = 0;
	while(s.length) {
		int index = rx.find(s, rx.RegExp(`\{[#$]`));
		if(index == -1) {
			// we have a normal block, then we're done
			Block* b = new Block;
			b.t = Block.Type.Normal;
			b.nb.contents = s[0..$];
			appendBlock(b);

			abspos += s.length;
			s = s[$..$];
		} else if(index == 0) {
			// we're on one now
			int end;
			string special;
			Block* b = handleTag(s, end, special);
			if(b is null) {
				if(terminating == 0)
					throw new Exception("template: bad end tag");
				if(terminating == 1)
					if(special == "else"){
						*ending = end + abspos;
						return start;
					} else if(special == "endif"){
						*ending = end + abspos;
						return start;
					}else
					throw new Exception("template: wrong tag");
				else if(terminating == 2)
					if(special == "endif") {
						*ending = end + abspos;
						return start;
					} else
					throw new Exception("template: wrong end");
				else if(terminating == 3)
					if(special == "endloop") {
						*ending = end + abspos;
						return start;
					} else
					throw new Exception("template: wrong end");
			} else {
				appendBlock(b);
				abspos += end;
				s = s[end..$];
			}
		} else {
			// we have a normal block, then the tag.
			Block* b = new Block;
			b.t = Block.Type.Normal;
			b.nb.contents = s[0..index];
			appendBlock(b);

			abspos += index;
			s = s[index..$];
		}
	}

	return start;
}

Block* handleTag(string t, out int end, out string special){
	invariant(int) index = indexOf(t, "}");
	if(index == -1) {
		// This isn't actually a tag at all!
		throw new Exception("Template error: tag not closed");
	} else {
		auto tag = t[1..index];
		if(tag[0] == '$') {
			Block* b = new Block;
			b.t = Block.Type.Variable;
			b.vb.variable = tag[1..$];

			end = index + 1;
			return b;
		}
		tag = tag[1..$];
		auto words = split(tag, " ");
		switch(words[0]) {
		  case "if":
		  	Block* b = new Block;
			b.t = Block.Type.If;
			b.ifb.variable = words[1];

			b.ifb.trueBlock = readTemplate(t[index+1..$], &end, 1);
			end += index + 1;
			if(t[end-1..end+5-1] == "#else"){ // FIXME: hackish
				int oldend = end;
				b.ifb.falseBlock = readTemplate(t[end+6..$], &end, 2);
				end += oldend + 6;
			}

			return b;
		  break;
		  case "nif":
		  	Block* b = new Block;
			b.t = Block.Type.If;
			b.ifb.variable = words[1];
			b.ifb.negated = true;


			b.ifb.trueBlock = readTemplate(t[index+1..$], &end, 1);
			end += index + 1;
			if(t[end-1..end+5-1] == "#else"){ // FIXME: hackish
				int oldend = end;
				b.ifb.falseBlock = readTemplate(t[end+6..$], &end, 2);
				end += oldend + 6;
			}


	

			return b;
		  break;
		  case "loop":
		  	Block* b = new Block;
			b.t = Block.Type.Loop;
			b.lb.variable = words[1];
			if(words[2] != "as")
				throw new Exception("template: missing 'as'");
			b.lb.as = words[3];

			b.lb.contents = readTemplate(t[index+1..$], &end, 3);
			end += index + 1;

			return b;
		  break;
		  case "else":
		  	end = -5; // put it back on top of #else
		  case "endif":
		  case "endloop":
			special = words[0];
			end += index + 1;
		  	return null;
		  case "include":
		  	Block* b = new Block;
			b.t = Block.Type.Include;
			auto str = cast(string) read(words[1]);
			Block* bb = readTemplate(str);
			b.ib.contents = bb;

			end += index + 1;

			// FIXME: set the special variables

			return b;
		  break;
//		  default:
			// not a tag at all again!
		}
	}

	assert(0);
}

	string outputBlocks(Block* b, TemplateVariable[string] v){
		string o = "";
		Block* current = b;
		int c = 0;
		while(current){
			switch(current.t){
			  case Block.Type.If:
			  	bool cond;
				cond = getVar(current.ifb.variable, v) !is null;
				if(current.ifb.negated)
					cond = !cond;
			  	if(cond)
					o ~= outputBlocks(current.ifb.trueBlock, v);
				else
					o ~= outputBlocks(current.ifb.falseBlock, v);
			  break;
			  case Block.Type.Loop:
			  	auto p = getVar(current.lb.variable, v);
				if(p !is null){
				TemplateVariable r = *p;
				if(r.t != TemplateVariable.Type.Array)
					throw new Exception("you ask the impossible!");
				// potential FIXME above; it might be possible

				foreach(a; r.arr){
					auto v2 = v;
					v2[current.lb.as] = a;
					o ~= outputBlocks(current.lb.contents, v2);

				}
				}
			  break;
			  case Block.Type.Include:
			  	// FIXME: add the include variables here
				o ~= outputBlocks(current.ib.contents, v);
			  break;
			  case Block.Type.Normal:
				o ~= current.nb.contents;
			  break;
			  case Block.Type.Variable:
			  	auto vr = getVar(current.vb.variable, v);
				if(vr is null)
					o ~= "(variable not set: " ~ current.vb.variable ~ ")";
				else
					o ~= vr.toString();
			  break;
			}

			current = current.next;
		}

		return o;
	}



	string output(string filename) {
		string a = cast(string) read(filename);
		Block* b = readTemplate(a);
		return outputBlocks(b, vars);
	}




	TemplateVariable* getVar(string v, TemplateVariable[string] vs){
	// FIXME: nulls
		auto words = v.split(".");
		if((words[0] in vs) is null)
			return null;
		auto res = &vs[words[0]];
		foreach(w; words[1..$]){
			if(res.t != TemplateVariable.Type.Hash)
				throw new Exception("Template error: type mismatch");
			auto lol = (w in res.hash);
			if(lol)
				res = lol;
			else
				throw new Exception("variable '"~w~"' not set in template; asking for: "~v);
		}

		return res;
	}

////////////////////
	TemplateVariable[string] vars;

	void add(string s, string v) {
		TemplateVariable a;
		a.t = TemplateVariable.Type.String;
		a.str = v;
		vars[s] = a;
	}

	void add(string s, string[] v) {
		TemplateVariable a;
		a.t = TemplateVariable.Type.Array;
		TemplateVariable b;
		foreach(d; v){
			b.t = TemplateVariable.Type.String;
			b.str = d;

			a.arr ~= b;
		}
		vars[s] = a;

	}

	void add(string s, string[string] vals) {
		TemplateVariable a;
		a.t = TemplateVariable.Type.Hash;
		foreach(k, v; vals) {
			TemplateVariable b;
			b.t = TemplateVariable.Type.String;
			b.str = v;
			a.hash[k] = b;
		}
		vars[s] = a;
	}

	void add(string s, string[string][] valz) {
		TemplateVariable t;
		t.t = TemplateVariable.Type.Array;
		foreach(values; valz) {
			TemplateVariable a;
			a.t = TemplateVariable.Type.Hash;
			foreach(k, v; values) {
				TemplateVariable b;
				b.t = TemplateVariable.Type.String;
				b.str = v;
				a.hash[k] = b;
			}


			t.arr ~= a;
		}

		vars[s] = t;
	}
}


// avatar, author, rank, contents, sig
import std.file;
import std.stdio;
/+

void main(){

	auto t = new Template;

	string[string] post;
	string[string][] posts;

	post["avatar"] = "/home/me/images/new/avatars/sdnav.jpg";
	post["author"] = "Adam D. Ruppe";
	post["rank"] = "Site Administrator";
	post["contents"] = "<p>Test post.</p>";
	post["sig"] = "<p>Nice sig!</p>";

	posts ~= post;


	string[string] post2;

	post2["avatar"] = "/home/me/images/new/avatars/smoonav.jpg";
	post2["author"] = "Sailor Moon";
	post2["rank"] = "Moon Princess";
	post2["contents"] = "<p>Test post #2.</p>";
	post2["sig"] = "<p>In the name of the moon!</p>";

	posts ~= post2;

	for(int a = 0; a< 4; a++){
	posts ~= post;
	posts ~= post2;
	}

	t.add("posts", posts);

	auto s = t.output("file.html");

	writefln("%s", s);

}
+/
