import std.string;
static import std.regexp;

class MarkedUpText {
	struct Element {
		ElementType* type; // null is just literal text or generic close tag ( [/] )

		string attribute; // contains the whole block of text if type is null

		bool isCloseTag;

		string originalText; // includes the opening [ and everything to the closing ]

		Element* next;
	}
/////////////////
	void readBBCode(string text) {
		Element* start;
		Element* current;

		void append(Element* e) {
			if(start is null)
				start = e;
			else
				current.next = e;
			current = e;
		}

		Element* plain(string c, int end) {
			Element* e = new Element;
			e.type = null;
			e.attribute = c[0..end];
			e.originalText = c[0..end];
			e.isCloseTag = false;

			return e;
		}

		int end;
		for(string c = text; c.length; c = text[0..end]) {
			int idx;
			idx = find(c, "[");
			if(idx != 0) {
				end = (idx == -1) ? c.length : idx;
				append(plain(c, end));
				continue;
			}

			// we're on some kind of tag, maybe
			string an = `[^\[\]]`; // describes the letters allowed as part of the article name
			auto r = std.regexp.RegExp(`\[\[(`~an~`*)\]\]`);
			int i = r.find(c); 
			if(i == 0) {
				// we have a wiki style link
				// FIXME
				continue;
			}
			auto r2 = std.regexp.RegExp(`\[\d+\]`);
			if(r2.find(c) == 0) {
				// we have a footnote reference
				// FIXME
				continue;
			}

			// Time to do some general extraction
			int ni1 = find(c, "=");
			int ni2 = find(c, "]");

			// FIXME: make sure the ] isn't inside a quoted attribute!
			// If it is, adjust ni2 appropriately.

			bool close = c[1] == '/';

			if(ni2 == -1) {
				// The ] is no where to be found; must be plain text.
				end = c.length;
				append(plain(c, end));
				continue;
			}

			if(c[0..ni2] == "[/]") {
				// we have a generic close tag
				Element* e = new Element;
				e.type = null;
				e.attribute = "";
				e.isCloseTag = true;
				e.originalText = c[0..ni2];

				end = ni2 + 1;

				append(e);
				continue;
			}

			string name;
			string attribute;

			if(ni1 > ni2 || ni1 == -1) {
				// the attribute isn't actually there - the = is past the ].
				name = c[(close ? 2 : 1)..ni2];
				attribute = "";
			} else { // we do indeed have an attribute
				name = c[(close ? 2 : 1)..ni1];
				attribute = c[ni1..ni2]; // note that it begins with the =
			}

			// We'll be using this a lot in a bit
			void invalid() {
				end = find(c[1..$], "[") + 1;
				if(end == -1)
					end = c.length;
				append(plain(c, end));
			}

			// Next up, look up the name
			ElementType* t = getElementType(name);

			if(t is null) {
				invalid();
				continue;
			}

			if(close && attribute.length) {
				// close tags aren't allowed to have attributes, so this
				// must be a text block...
				invalid();
				continue;
			}

			if(close && !t.closes) {
				// This tag isn't meant to close
				invalid();
				continue;
			}

			if(t.attribute == None && attribute.length) {
				// Tag shouldn't have an attribute
				invalid();
				continue;
			}

			if(t.attributeRequired && attribute.length == 0) {
				// Tag *should* have an attribute
				invalid();
				continue;
			}

			// FIXME: do a syntax check on the attribute

			// If we've made it this far, it must be valid.

			end = ni2 + 1;

			Element* e;
			e.type = t;
			e.attribute = attribute;
			e.isCloseTag = close;
			e.originalText = c[0..end];
			append(e);
		}

		input = start;
	}

	Element* input;

	void parseInput() {
		for(Element* e = input; e !is null; e = e.next) {

		}
	}

	struct Part {
		Element* thisOne;

		Part* contents;

		Part* next;
	}

	Part readPart(inout Element* e, int recursion = 0) {
		if(e.type is null) {
			auto t = new Text;
		} else switch(e.type.name) {
		  case "quote":
		  break;
		  case "table":
		  break;
		  case "list":
		  break;
//		  case "p":
//		  case "block":
		  default:
		}
	}
}

// ///////////////////////////


ElementType* getElementType(string name) {
	foreach(i, t; allElements)
		if(t.name == name)
			return &(allElements[i]);
	return null;
}

enum { None, Text, NestedTags, Color, Number, Percentage }
alias int AttributeType;

struct ElementType {
	string name;
	bool closes;
	AttributeType attribute;
	bool attributeRequired;
	string meaning;
}

ElementType[] allElements = [
	{"quote", true, NestedTags, false, "Quotation"},
	{"code", true, Text, false, "Code"},
	{"table", true, Text, false, "Data table"},
	{"table2", true, Text, false, "Data table inverted"},
	{"header", false, None, false, "Table headers"},
	{"body", false, None, false, "Table body"},
	{"footer", false, None, false, "Table footer data"},
	{"url", true, Text, false, "Hyperlink"},
	{"img", true, Text, false, "Inline image"},
	{"list", true, Text, false, "List"},
	{"pre", true, None, false, "Preformatted text"},
	{"p", true, Text, false, "Paragraph - attribute is indent or no"},
	{"br", false, None, false, "Forced line break"},
	{"b", true, None, false, "Bold"},
	{"i", true, None, false, "Italics"},
	{"u", true, None, false, "Underline"},
	{"tt", true, None, false, "Monospace"},
	{"sup", true, None, false, "Superscript"},
	{"sub", true, None, false, "Subscript"},
	{"q", true, NestedTags, false, "Brief quotation"},
	{"color", true, Color, true, "Text color"},
	{"size", true, Number, true, "Text size"},
	{"highlight", false, None, false, "Highlighted text"},
	{"hidden", false, None, false, "Hidden (white on white) text"},
	{"big", true, None, false, "Big text"},
	{"small", true, None, false, "Small text"},
	{"spoiler", true, Text, false, "Spoiler block"},
	{"article", true, Text, false, "Link to an article on the site"}, // alias [[attribute|text]]
	{"*", false, None, false, "Marks an element of a list or table"},
	{"text", true, None, false, "Marks plain text"},


	{"section", true, Text, true, "Marks a section of the article"},
	{"align", true, Text, true, "Text alignment"},
	{"float", true, Text, true, "Make a floating box"},
	{"abbr", true, Text, true, "Defines an abbreviation"},
	{"hr", false, Number, false, "Draws a line"},
	{"footnote", true, Number, false, "Defines a footnote"},
	{"dictionary", true, None, false, "Defines a glossary section"},
	{"toc", false, None, false, "Outputs a table of contents of your sections"},
	{"footnotes", false, None, false, "Outputs all your footnotes"},
	{"fn", false, Number, true, "Outputs a link to a footnote"}, // alias [attribute] iff footnote defined
	{"block", true, Number, false, "Outputs a centered, blocked paragraph"},


	{"math", true, None, false, "Outputs rendered math notation"},
	{"equation", true, Number, false, "Outputs rendered equation"},
	{"graph", true, Text, false, "Outputs a graph of the data"},
	{"html", true, None, false, "Outputs the given html unmodified"}
];
/*
bhere is some bold text/b
[b]here is some bold text[/b]
<b>here is some bold text</b>

[tag]
[##] shorthand for footnote
[@glossary item]
[.user defined]
[[link]]

*/
// First, we'll convert shorthand to full tags
// Then we read the elements.
