import program;
import std.stdio;

static import serverconnection;

Display[int] displays;

class Display : serverconnection.GeneratedConnection {
	static int lastDisplayId;
	int did;
	override void onConnect() {
		did = ++lastDisplayId;
		displays[did] = this;
		foreach(c; programs) {
			newClient(c.pid);
			if(c.name)
			clientNameChanged(c.pid, c.name);
//			clientAttachedElsewhere(int id);
//			clientDetached(int id);
//			clientTerminated(int id);
		}

	}

	override void onDisconnect() {
		displays.remove(did);
		foreach(c; connectedTo) {
			c.display = null;
			c.detach();
		}
	}

	Program[int] connectedTo;


/*
	Events can carry one of these IDs:
		Window id
		Control id
		Action id



	ID Mapping goes:
		program reference -> its own Id
*/


	override void windowCloseRequested(int id) {
		foreach(p; connectedTo) {
			auto a = p.windows.getCidFromId(id);
			if(a)
				p.windowCloseRequested(a);
		}

	}

	override void textChanged(int id, string text) {
		foreach(p; connectedTo) {
			auto a = p.controls.getCidFromId(id);
			if(a) {
				p.textChanged(a, text);
				p.controls.get(a).text = text;
			}
		}

	}

	override void textInserted(int id, int pos, string text) {
		foreach(p; connectedTo) {
			auto a = p.controls.getCidFromId(id);
			if(a) {
				p.textInserted(a, pos, text);

				string old = p.controls.get(a).text;
				if(pos > old.length)
					pos = old.length;
				p.controls.get(a).text = old[0..pos] ~ text ~ old[pos..$];
			}
		}

	}


	override void textRemoved(int id, int pos, int count) {
		foreach(p; connectedTo) {
			auto a = p.controls.getCidFromId(id);
			if(a) {
				p.textRemoved(a, pos, count);

				string old = p.controls.get(a).text;
				p.controls.get(a).text = old[0..pos] ~ old[pos+count..$];
			}
		}

	}


	override void actionTriggered(int id) {
		foreach(p; connectedTo) {
			auto a = p.actions.getCidFromId(id);
			if(a)
				p.actionTriggered(a);
		}
	}
	
	override void attachClient(int id) {
		if( auto c = id in programs ) {
			c.connectToDisplay(this);
			connectedTo[id] = *c;
		}

	}

	override void detachClient(int id) {
		if( auto c = id in programs )
			if(c.display is this) {
				c.detach();
				connectedTo.remove(id);
			}
	}

	override void controlTriggered(int id) {
		foreach(p; connectedTo) {
			auto a = p.controls.getCidFromId(id);
			if(a)
				p.controlTriggered(a);
		}
	}

	public immutable int id;
	this() { id = ++lastId; }
	private static int lastId;
}
