import display;

static import clientconnection;

Program[int] programs;




/+
struct StuffToDo {
	void gotId(int id) {
		foreach(thing; things)
			thing(id);
	}

	void add(void delegate(int) thing) {
		things ~= thing;
	}

	void delegate(int id) things[];
}
+/






/*
	Reconnect procedure:

	Create the item in its present state
		that's easy - running status does its id
	
	Then we need to put the items in their containers and such
		We need the ID for this, so we're probably gonna have to wait
		for the next packet.

	So we create it and store the ID we have here
	Use running status to update its own state
	
	When updating state, if running status isn't the same item
	we just worked with, we need to append it to the delegate
	to take care of it as soon as we know the ID.
*/


/*
	IDs: the one on the server and the one on the manager/client

	So now items have two IDs: the one the client expects and the one we have (and dictate to the server).


	So here we store the real current state of the item
	When it changes we notify the display to send the update
	When an event comes in about it or a response, the server tells us to update this list

*/

struct Action {
	int id;
	string text;

	void reCreate(Display display) {
		create(display);
		display.setActionText(id, text);
	}

	void create(Display display) {
		display.createAction(id);
	}

	void destroy(Display display) {

	}
}

struct Window {
	int id;
	string title;
	string statusBarText;
	bool showing;

	void reCreate(Display display) {

		create(display);
		display.setWindowTitle(id, title);

		foreach(menu; menus) {
			display.addMenu(id, menu.id);
		}
/*
		foreach(control; controls) {
			control.replace(display,&this);
		}
*/
	}

	void create(Display display) {
		display.createMainWindow(id);
	}

	void destroy(Display display) {
		display.closeWindow(id);

	}

	void addMenu(Display display, Menu* menu) {
		menus ~= menu;
		if(display)
		display.addMenu(id, menu.id);
	}

	Menu*[] menus;

//	Control*[] controls
}

struct Menu {
	int id;
	string title;

	void reCreate(Display display) {
		create(display);
		foreach(i; items)
			i.create(display);
	}

	void create(Display display) {
		display.createMenu(id, title);
	}

	void destroy(Display display) {

	}

	void addItem(Display d, Action* a) {
		Item i;
		i.isSeparator = false;
		i.action = a;
		i.pid = id;

		items ~= i;

		i.create(d);
	}

	Item[] items;

	struct Item {
		bool isSeparator;
		// the following are only if !separator
		Action* action;
		int pid; // of hte parent

		void create(Display display) {
			if(!display)
				return;
			if(isSeparator)
				display.addMenuSeparator(pid);
			else
				display.addMenuItem(pid, action.id);

		}
	}
}


struct Console {
	int id;

	void create(Display display) {
		display.createConsole(id);
	}

	void reCreate(Display display) {
		create(display);
	}

	void destroy(Display display) {

	}
}


struct Control {
	int id;
	int type;
	int parent;

	int x;
	int y;
	int w;
	int h;

	int scrollPosition; // where the scrollbar is on a text viewer, etc
	int value;
	int value2;
	int maxValue;
	int minValue;
	int imagePosition;
	string text;
	int[] items; // should contain flags for selected, etc

	bool disabled;
	bool hidden;

	void create(Display display) {
		display.createControl(id, type, parent);
	}

	void reCreate(Display display) {
		create(display);
		setGeometry(display, x, y, w, h);
		if(text != "")
			display.setControlText(id, text);
	}

	void destroy(Display display) {

	}

	void setGeometry(Display display, int x, int y, int w, int h) {
		this.x = x;
		this.y = y;
		this.w = w;
		this.h = h;

		if(display)
		display.setControlGeometry(id, x, y, w, h);
	}
}

struct Resource(T) {
	T* create(int clientGivenId, Display display, void delegate(T*) init = null) {
		lastId++;
		T* t = new T;
		t.id = lastId;

		all[t.id] = t;

		mapping[clientGivenId] = t.id;

		if(init)
			init(t);

		if(display)
			t.create(display);

		return t;
	}

	void destroyAll(Display display) {
		foreach(lol, a; all) {
			if(lol == 0) continue;
				a.destroy(display);
		}
	}

	void recreateAll(Display display) {
		foreach(lol, a; all) {
			if(lol == 0) continue;
				a.reCreate(display);
		}
	}

	T* get(int cid) {
		return all[mapping[cid]];
	}

	int getCidFromId(int id) {
		foreach(i, a; mapping)
			if(a == id)
				return i;
		return 0;
	}

	T*[int] all; // [Global ID] = this one
	int[int] mapping; // [ID the client client expects] = Global ID

	static int lastId;
}

import std.stdio;

class Program : clientconnection.GeneratedConnection {
	void connectToDisplay(Display d) {
	/*
		if(d is display)
			return; // we are already attached there
	*/

		// If we are attached somewhere, we first need to detach from that location
		detach();

		display = d;
		if(d is null)
			return;
		// We need to tell the display to recreate all the
		// resources the program is using and tell the program
		// that things have changed

		// The order in which they are recreated is important - anything that might be referenced later needs to be
		// created first. Thus we do things without children, then things that might have children.

		actions.recreateAll(display);

		menus.recreateAll(display);

		consoles.recreateAll(display);

		windows.recreateAll(display);
		controls.recreateAll(display);

		/* we show them all at the end, after they are already reconstructed */
		foreach(window; windows.all)
			if(window.showing)
				display.showWindow(window.id);
	}

	void detach() {
		if(display !is null) {
			/* we hide all the windows so their destruction appears instant */
			foreach(window; windows.all)
				if(window.showing)
					display.hideWindow(window.id);
			controls.destroyAll(display);
			windows.destroyAll(display);
			consoles.destroyAll(display);
			menus.destroyAll(display);
			actions.destroyAll(display);
		}
		display = null;
	}

	override void onConnect() {
		lastProgramId++;
		pid = lastProgramId;
		programs[pid] = this;
		foreach(d; displays)
			d.newClient(pid);
	}

	override void onDisconnect() {
		programs.remove(pid);
		foreach(d; displays)
			d.clientTerminated(pid);
		detach();
	}



	// internal data
	private static int lastProgramId;
	Display display;
	int pid;

	string name;

	// All the resource handles the client holds
	Resource!(Window) windows;
	Resource!(Menu) menus;
	Resource!(Action) actions;
	Resource!(Console) consoles;
	Resource!(Control) controls;

	// Implementation of the actual calls the client can make
	// When we can, we execute them here as well as passing them
	// on to the actual display server.


	override void setApplicationTitle(string newname) {
		name = newname;
		foreach(d; displays)
			d.clientNameChanged(pid, name);

	}

	override void closeWindow(int id) {
		Window* w = windows.get(id);
		w.destroy(display);
		windows.mapping.remove(id);
		windows.all.remove(w.id);

		int[] ids;
		foreach(c; controls.all) {
			if(c.parent == w.id) {
				c.destroy(display);
				controls.mapping.remove(id);
				ids ~= c.id;
			}
		}
		foreach(i; ids)
			controls.all.remove(i);
	}

	override void appendControlText(int id, string text) {
		controls.get(id).text ~= text;
		if(display)
			display.appendControlText(controls.get(id).id, text);
	}

	override void createControl(int id, int type, int parent) {
		Control* n = controls.create(id, display, (Control* t) { t.type = type; t.parent = windows.get(parent).id; });
	}

	override void setControlGeometry(int id, int x, int y, int w, int h) {
		controls.get(id).setGeometry(display, x, y, w, h);
	}

	override void setControlText(int id, string text){
		controls.get(id).text = text;
		if(display)
			display.setControlText(controls.get(id).id, text);
	}

	override void createConsole(int id) {
		Console* n = consoles.create(id, display);
	}

	override void createMainWindow (int val){
		Window* n = windows.create(val, display);
	}

	override void createMenu(int val, string title) {
		Menu* n = menus.create(val, display);
		n.title = title;
	}

	override void addMenuItem(int menuId, int actionId) {
		menus.get(menuId).addItem(display, actions.get(actionId));
	}

	override void addMenu(int wid, int menuId) {
		windows.get(wid).addMenu(display, menus.get(menuId));
	}

	override void createAction(int id) {
		Action* a = actions.create(id, display);
	}

	override void setActionText(int id, string text) {
		actions.get(id).text = text;
		if(display)
			display.setActionText(actions.get(id).id, text);
	}

	override void messageBox (string title, string message){
		if(display)
		display.messageBox(title, message);
	}


	override void setWindowTitle (int wid, string title){
		windows.get(wid).title = title;
		if(display)
		display.setWindowTitle(windows.mapping[wid], title);
	}

	override void showWindow(int wid) {
		windows.get(wid).showing = true;
		if(display)
		display.showWindow(windows.mapping[wid]);
	}


}
