import std.stdio;
public import arsd.netman;

class DWSApp {
  public:
	this(string name = null) {
		el = new EventLoop;
		d = openDisplay(el);

		if(name !is null)
		d.setApplicationTitle(name);

		singleton = this;
	}
	
	void exec() {	
		while( el.proceed( ) ) { }
	}

	void setConnectionSpawner(int port, Connection delegate() c) {
		assert(c !is null);
		el.allocs[port] = c;
	}

	Display d;
	EventLoop el;

	static DWSApp singleton;
}


// /////////////////////////

// private
template ServerResource(T) {
	public void init(Display d = null) {
		if(d is null)
			display = Display.newestDisplay;
		else
			display = d;
		assert(display !is null);
		id = ++ lastId;

		all[id] = this;
	}

	public int id;
	private static int lastId;
	protected Display display;
	private static T[int] all;
}



class Action {
	string text;
	Image icon;

	this(string text, void delegate() t, Display d = null) {
		init(d);
		triggered = t;

		display.createAction(id);
		display.setActionText(id, text);
	}


	void delegate() triggered;
	mixin ServerResource!(Action);
}

abstract class Control {

	this(Window p, Display d) {
		init(d);
		parent = p;
	}

	void setGeometry(int x, int y, int w, int h) {
		display.setControlGeometry(id, x, y, w, h);
	}

	Window parent;



	string text;

	int cursorLocation;
	int selectionStart;
	int selectionEnd;

	int value;
	int value2;
	int max;
	int min;

	void delegate() onPress;
	mixin ServerResource!(Control);
}

abstract class Window {

	string title;


	void show() {
		display.showWindow(id);
	}

	// returning true means go ahead and close
	bool delegate() onClose;

	void closeEvent() {
		if(onClose) {
			if(onClose()) {
				close();
			}
		} else
			close();
	}

	void close() {
		display.closeWindow(id);
		windowsOpen--;

		if(windowsOpen == 0) {
			throw new Exception("all windows closed"); // FIXME
		}
	}

	private static int windowsOpen;
	mixin ServerResource!(Window);
}

class Menu {
	this(string title, Display d = null) {
		init(d);

		display.createMenu(id, title);
	}
	
	void addSeparator() {
		display.addMenuSeparator(id);
	}

	void addItem(Action act) {
		display.addMenuItem(id, act.id);
	}

	mixin ServerResource!(Menu);
}

class Item {
	string name;
	Image icon;

	mixin ServerResource!(Item);
}

class DataStore {

	mixin ServerResource!(DataStore);
}

class Image {

	mixin ServerResource!(Image);
}


// ////////////////////////

Display openDisplay(NetworkManager netman, string host = "localhost", ushort port = 8081) {
	netman.connect(host, port);

	return Display.newestDisplay;
}

class EventLoop : NetworkManager {
	override Connection allocConnection(int port) {
		if(port == 8081)
			return new Display();
		if(auto a = port in allocs)
			return (*a)();
		assert(0);
	}

	Connection delegate()[int] allocs;
}


// //////////////////

import gen;

class Display : GeneratedConnection {
	override void onConnect() {
		newestDisplay = this;
	}

	override void onDisconnect() {
		throw new Exception("display connection lost " ~ reason);
	}

	override void controlTriggered(int id) {
		auto a = id in Control.all;
		if(a !is null)
			a.onPress();
	}

	override void actionTriggered(int id) {
		auto a = id in Action.all;
		if(a is null)
			writefln("wtf %d", id);
		else
			if(a.triggered)
			a.triggered();
	}

	override void windowCloseRequested(int id) {
		auto a = id in Window.all;
		if(a is null)
			writefln("no window %d", id);
		else
			a.closeEvent();
	}


	override void textChanged(int id, string text) {
		auto a = id in Control.all;
		if(a)
			a.text = text;
	}

	override void textInserted(int id, int pos, string text) {
		auto a = id in Control.all;
		if(a) {
			string old = a.text;
			if(pos > old.length)
				pos = old.length;
			a.text = old[0..pos] ~ text ~ old[pos..$];
		}

	}


	override void textRemoved(int id, int pos, int count) {
		auto a = id in Control.all;
		if(a) {

			string old = a.text;
			a.text = old[0..pos] ~ old[pos+count..$];
		}
	}


	private static Display newestDisplay;
}


